# Copyright (C) 2001-2011, Christof Meerwald
#               2018, Jens W. Wulf
# http://jabrss.cmeerw.org, http://www.jotwewe.de/de/xmpp/jabrss

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

from __future__ import with_statement

import bisect, threading, ssl, socket, os
import sys, logging, time, traceback

from xmpplify import tobytes, Element, JID, Stanza, XmppStream
from parserss import RSS_Resource, RSS_Resource_id2url, RSS_Resource_simplify
from parserss import RSS_Resource_db
from parserss import UrlError

import FlexibleLocker
import Cursor
import Misc

import globals
import messages

MAX_MESSAGE_SIZE = 20000

class JabRSSStream(XmppStream):
    def __init__(self, jid, host, password, port=5222):
        self.jid, self._host, self._port = jid, host, port
        self._update_queue = []
        self._update_queue_cond = threading.Condition()
        RSS_Resource.schedule_update = self.schedule_update

        self._io_sync = threading.Lock()
        self._roster_id, self._sock, self._encoding = 0, None, 'utf-8'
        self._online = False
        self._term, self._term_flag = threading.Event(), False

        handlers = {
            ('iq', 'get', '{http://jabber.org/protocol/disco#info}query') : self.iq_get_disco,
            ('iq', 'get', '{jabber:iq:last}query') : self.iq_get_last,
            ('iq', 'get', '{urn:xmpp:ping}ping') : self.iq_get_ping,
            ('iq', 'get', '{jabber:iq:time}query') : self.iq_get_time_jabber,
            ('iq', 'get', '{urn:xmpp:time}time') : self.iq_get_time,
            ('iq', 'get', '{jabber:iq:version}query') : self.iq_get_version,
            ('iq', 'get') : self.iq_get,
            ('iq', 'set', '{jabber:iq:roster}query') : self.iq_set_roster,
            ('iq', 'set') : self.iq_set,
            ('message', 'normal') : self.message,
            ('message', 'chat') : self.message,
            ('message', 'headline') : self.message_headline,
            ('presence', None) : self.presence_available,
            ('presence', 'unavailable') : self.presence_unavailable,
            ('presence', 'error') : self.presence_unavailable,
            ('presence', 'subscribe') : self.presence_subscribe,
            ('presence', 'subscribed') : self.presence_subscribed,
            ('presence', 'unsubscribe') : self.presence_unsubscribe,
            ('presence', 'unsubscribed') : self.presence_unsubscribed,
            }
        XmppStream.__init__(self, self.jid, handlers,
                            encoding=self._encoding,
                            password=password)


    def _stream_closed(self):
        globals.storage.evict_all_users()
        sock = self._sock
        self._sock, self._online = None, False
        return sock


    def connect(self):
        # resolve addresses manually for compatibility with Python 2.5
        # which doesn't support create_connection
        exc = socket.gaierror(-2, 'Name or service not known')
        for ai in socket.getaddrinfo(self._host, self._port, socket.AF_UNSPEC, socket.SOCK_STREAM):
            try:
                exc = None
                self._sock = socket.socket(ai[0], ai[1], ai[2])
                self._sock.settimeout(30)
                self._sock.connect(ai[4])
                break
            except socket.error as e:
                exc = e

        if exc != None:
            raise exc

        globals.log_message_info('connected to', str(self._sock.getpeername()))
        self._sock.settimeout(600)
        XmppStream.connect(self)


    def sock(self):
        self._io_sync.acquire()
        try:
            return self._sock
        finally:
            self._io_sync.release()

    def send(self, data):
        self._io_sync.acquire()
        try:
            if self._sock != None:
                globals.logger.debug('>>> ' + repr(data))
                try:
                    self._sock.sendall(data)
                except socket.error:
                    self._stream_closed()
                    raise
        finally:
            self._io_sync.release()

    def closed(self):
        globals.log_message_info('stream closed')
        self._io_sync.acquire()
        try:
            self._stream_closed()
        finally:
            self._io_sync.release()

    def shutdown(self):
        globals.log_message_info('stream shutdown')
        self._io_sync.acquire()
        try:
            sock = self._stream_closed()
            if sock != None:
                try:
                    sock.shutdown(socket.SHUT_WR)
                except socket.error:
                    pass
        finally:
            self._io_sync.release()


    def terminate(self):
        self._term_flag = True

        self._update_queue_cond.acquire()
        self._update_queue_cond.notifyAll()
        self._update_queue_cond.release()
        self._term.wait()

    def terminated(self):
        return self._term_flag


    def stream_start(self, elem):
        globals.log_message_info('stream start')

    def stream_features(self, elem):
        globals.log_message_info('stream features')

    def stream_error(self, elem):
        globals.log_message_err('stream error')
        globals.storage.evict_all_users()

    def stream_end(self, elem):
        globals.storage.evict_all_users()


    def starttls_proceed(self, elem):
        globals.log_message_info('starttls proceed')
        self._sock = ssl.wrap_socket(self._sock, do_handshake_on_connect=False)
        self._sock.do_handshake()
        XmppStream.connect(self)

    def starttls_failure(self, elem):
        # we are unable to recover here...
        globals.log_message_err('starttls failure')
        self._term_flag = True
        self.disconnect()


    def sasl_failure(self, elem):
        # we are unable to recover from these...
        globals.log_message_err("authentication failed")
        self._term_flag = True
        self.disconnect()

    def sasl_success(self, elem):
        globals.log_message_info('authenticated')


    def iq_get(self, iq):
        logmsg = ['iq get']
        iq_id, iq_from = iq.get_id(), iq.get_from()
        if iq_id != None:
            logmsg.append(iq_id)
        if iq_from != None:
            logmsg.append(iq_from.tostring())
        query = iq.get_query()
        if query != None:
            logmsg.append(query.tag)
        globals.log_message_info(' '.join(logmsg))

        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='error',
                              from_=iq.get_to(), to=iq_from)
            error = reply.create_error(type='cancel')
            error.append(Element('{urn:ietf:params:xml:ns:xmpp-stanzas}service-unavailable'))
            self.send(reply.asbytes(self._encoding))

    def iq_get_disco(self, iq):
        globals.log_message_info('iq get disco', iq.get_from().tostring())
        iq_id = iq.get_id()
        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='result',
                              from_=iq.get_to(), to=iq.get_from())
            disco = reply.create_query('http://jabber.org/protocol/disco#info')

            identity = Element('{http://jabber.org/protocol/disco#info}identity')
            identity.set('category', 'client')
            identity.set('type', 'bot')
            disco.append(identity)

            for fvar in ('http://jabber.org/protocol/disco#info',
                         'jabber:iq:last', 'jabber:iq:time',
                         'jabber:iq:version',
                         'urn:xmpp:ping', 'urn:xmpp:time'):
                feature = Element('{http://jabber.org/protocol/disco#info}feature')
                feature.set('var', fvar)
                disco.append(feature)
            self.send(reply.asbytes(self._encoding))

    def iq_get_last(self, iq):
        globals.log_message_info('iq get last', iq.get_from().tostring())
        iq_id = iq.get_id()
        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='result',
                              from_=iq.get_to(), to=iq.get_from())
            last = reply.create_query('jabber:iq:last')
            last.set('seconds', '0')
            self.send(reply.asbytes(self._encoding))

    def iq_get_ping(self, iq):
        globals.log_message_info('iq get ping', iq.get_from().tostring())
        iq_id = iq.get_id()
        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='result',
                              from_=iq.get_to(), to=iq.get_from())
            self.send(reply.asbytes(self._encoding))

    def iq_get_time_jabber(self, iq):
        globals.log_message_info('iq get time', iq.get_from().tostring())
        iq_id = iq.get_id()
        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='result',
                              from_=iq.get_to(), to=iq.get_from())
            time_elem = reply.create_query('jabber:iq:time')
            utc = Element('{jabber:iq:time}utc')
            utc.text = time.strftime('%Y%m%dT%H:%M:%S', time.gmtime())
            time_elem.append(utc)
            self.send(reply.asbytes(self._encoding))

    def iq_get_time(self, iq):
        globals.log_message_info('iq get time', iq.get_from().tostring())
        iq_id = iq.get_id()
        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='result',
                              from_=iq.get_to(), to=iq.get_from())
            time_elem = reply.create_child('urn:xmpp:time', 'time')
            tzo = Element('{urn:xmpp:time}tzo')
            tzo.text = '-00:00'
            time_elem.append(tzo)
            utc = Element('{urn:xmpp:time}utc')
            utc.text = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
            time_elem.append(utc)
            self.send(reply.asbytes(self._encoding))

    def iq_get_version(self, iq):
        globals.log_message_info('iq get version', iq.get_from().tostring())
        iq_id = iq.get_id()
        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='result',
                              from_=iq.get_to(), to=iq.get_from())
            version = reply.create_query('jabber:iq:version')
            version_name = Element('{jabber:iq:version}name')
            version_name.text = 'JabRSS'
            version.append(version_name)
            version_version = Element('{jabber:iq:version}version')
            version_version.text = '0.x'
            version.append(version_version)
            version_os = Element('{jabber:iq:version}os')
            version_os.text = 'Linux'
            version.append(version_os)
            self.send(reply.asbytes(self._encoding))


    def iq_set(self, iq):
        logmsg = ['iq set']
        iq_id, iq_from = iq.get_id(), iq.get_from()
        if iq_id != None:
            logmsg.append(iq_id)
        if iq_from != None:
            logmsg.append(iq_from.tostring())
        query = iq.get_query()
        if query != None:
            logmsg.append(query.tag)
        globals.log_message_info(' '.join(logmsg))

        if iq_id != None:
            reply = Stanza.Iq(id=iq_id, type='error',
                              from_=iq.get_to(), to=iq_from)
            error = reply.create_error(type='cancel')
            error.append(Element('{urn:ietf:params:xml:ns:xmpp-stanzas}service-unavailable'))
            self.send(reply.asbytes(self._encoding))

    def iq_set_roster(self, iq):
        iq_id, query = iq.get_id(), iq.get_query()
        item = query.find('{jabber:iq:roster}item')
        if item != None:
            self.roster_updated(item)
            if iq_id != None:
                reply = Stanza.Iq(id=iq_id, type='result',
                                  from_=iq.get_to(), to=iq.get_from())
                self.send(reply.asbytes(self._encoding))

    def session_start(self):
        globals.log_message_info('session start')
        iq = Stanza.Iq(type='get', id='roster')
        iq.create_query('jabber:iq:roster')
        self.send(iq.asbytes(self._encoding))

        result = yield 'roster'
        globals.log_message_info('roster retrieved')
        query = result.get_query()
        self.roster_updated(query.getchildren())

        globals.log_message_info('sending presence')
        presence = Stanza.Presence()
        self.send(presence.asbytes(self._encoding))
        self._online = True
        return

    def update_presence(self):
        if self._online:
            presence = Stanza.Presence()

            total_users, total_resources = 0, 0
            with Cursor.Cursor(globals.db) as cursor:
                result = cursor.execute('SELECT (SELECT COUNT(uid) FROM user), (SELECT COUNT(DISTINCT rid) FROM user_resource)')
                for total_users, total_resources in result:
                    pass

            presence.set_status('%d/%d users, %d/%d feeds' %
                                (len(globals.storage._users) // 2,
                                 total_users,
                                 len(globals.storage._resources) // 2,
                                 total_resources))
            self.send(presence.asbytes(self._encoding))


    def unhandled_stanza(self, stanza):
        globals.log_message_err('unhandled stanza', stanza.tag())


    def _process_list(self, user):
        reply_body = []
        for res_id in user.resources():
            resource = globals.storage.get_resource_by_id(res_id, globals.main_res_db)
            res_updated, res_modified, res_invalid = resource.times()
            if res_invalid == None:
                reply_body.append(resource.url())
                for strLine in user.get_filter_strings(resource.id()):
                    reply_body.append('  ' + strLine)
            else:
                error_info = resource.error_info()
                if error_info:
                    reply_body.append('%s (Error: %s)' % (resource.url(),
                                                          error_info))
                else:
                    reply_body.append('%s (error)' % (resource.url(),))

        if len(reply_body) > 0:
            # no, because of filters: reply_body.sort()
            return ['\n'.join(reply_body)]
        else:
            return ['Sorry, you are currently not subscribed to any RSS feeds.']

    def _parse_format(self, args):
        format = 0
        for arg in args:
            if arg.lower() == 'title':
                format |= 1
            elif arg.lower() == 'url' or arg.lower() == 'link':
                format |= 2
            elif arg.lower() == '<empty>':
                format = 0
                break
            else:
                raise Exception('invalid format')
        return format

    def _process_set(self, user, argstr):
        try:
            arg = argstr.strip()
            if arg == 'plaintext':
                user.set_message_type(0)
                reply_body = 'Message type set to "plaintext"'
            elif arg == 'headline':
                user.set_message_type(1)
                reply_body = 'Message type set to "headline"'
            elif arg == 'chat':
                user.set_message_type(2)
                reply_body = 'Message type set to "chat"'
            else:
                args = arg.split()
                if args[0] == 'also_deliver':
                    deliver_cfg = 0

                    for s in args[1:]:
                        s = s.lower()
                        if s == 'away':
                            deliver_cfg = deliver_cfg | 1
                        elif s == 'xa':
                            deliver_cfg = deliver_cfg | 2
                        elif s == 'dnd':
                            deliver_cfg = deliver_cfg | 4
                        elif s == 'none':
                            pass
                        else:
                            raise Exception('unknown setting for "also_deliver"')

                    user.set_delivery_state(deliver_cfg)
                    reply_body = '"also_deliver" setting adjusted'
                elif args[0] == 'store_messages':
                    store_messages = int(args[1])
                    user.set_store_messages(store_messages)
                    reply_body = '"store_messages" setting adjusted'
                elif args[0] == 'size_limit':
                    size_limit = int(args[1])
                    user.set_size_limit(size_limit)
                    reply_body = '"size_limit" setting adjusted'
                elif args[0] == 'header':
                    format = self._parse_format(args[1:])
                    user.set_header_format(format)
                    reply_body = 'header format adjusted'
                elif args[0] == 'subject':
                    format = self._parse_format(args[1:])
                    user.set_subject_format(format)
                    reply_body = 'subject format adjusted'
                elif args[0] == 'filter' and len(args) > 2:
                    pos = arg.find("\n")
                    if pos < 0:
                        reply_body = messages.TEXT_HELP_FILTER
                    else:
                        reply_body = self._process_set_filter(user, args[1].strip(), arg[pos+1:])
                else:
                    reply_body = 'Unknown configuration option'
        except:
            reply_body = 'Unknown error setting configuration option'
        return [reply_body]


    def _format_format_conf(self, format):
        format_text = []
        if format & 1:
            format_text.append('title')
        if format & 2:
            format_text.append('url')
        if format_text == []:
            format_text.append('<empty>')

        return ', '.join(format_text)

    def _process_config(self, user):
        reply_body = ['Current configuration:']

        message_type = user.get_message_type()
        if message_type == 0:
            reply_body.append('message type "plaintext"')
        elif message_type == 1:
            reply_body.append('message type "headline"')
        elif message_type == 2:
            reply_body.append('message type "chat"')
        else:
            reply_body.append('message type <reserved>')

        deliver_when_away = user.get_deliver_when_away()
        deliver_when_xa = user.get_deliver_when_xa()
        deliver_when_dnd = user.get_deliver_when_dnd()
        if deliver_when_away or deliver_when_xa or deliver_when_dnd:
            deliver_list = []
            if deliver_when_away:
                deliver_list.append('Away')
            if deliver_when_xa:
                deliver_list.append('XA')
            if deliver_when_dnd:
                deliver_list.append('DND')
            reply_body.append('Headlines will also be delivered when you are %s' % (', '.join(deliver_list)))

        subject_format = user.get_subject_format()
        reply_body.append('subject format: %s' % (self._format_format_conf(subject_format),))

        header_format = user.get_header_format()
        reply_body.append('header format: %s' % (self._format_format_conf(header_format),))

        store_messages = user.get_store_messages()
        reply_body.append('At most %d headlines will be stored for later delivery' % (store_messages,))

        size_limit = user.get_size_limit()
        if size_limit:
            reply_body.append('The size of a headline message will be limited to about %d bytes' % (size_limit,))

        return ['\n'.join(reply_body)] # to be send as one message


    def _process_statistics(self, user):
        reply_body = ['Statistics:']
        total_users, total_resources = 0, 0
        with Cursor.Cursor(globals.db) as cursor:
            result = cursor.execute('SELECT (SELECT COUNT(uid) FROM user), (SELECT COUNT(DISTINCT rid) FROM user_resource)')
            for total_users, total_resources in result:
                pass

        reply_body.append('Users online/total: %d/%d' %
                          (len(globals.storage._users) // 2, total_users))
        reply_body.append('RDF feeds used/total: %d/%d' %
                          (len(globals.storage._resources) // 2, total_resources))

        return ['\n'.join(reply_body)] # to be send as one message


    def _process_usage(self, user):
        reply_body = ['Usage Statistics:']

        reply_body.append('subscribed to %d feeds' % (len(user.resources())))

        stat_start, nr_headlines, size_headlines = user.get_statistics()
        stat_start = stat_start - (len(nr_headlines) - 1)

        time_base = stat_start * 7*24*60*60 - 60*60*60

        for i in range(0, len(nr_headlines)):
            nr = nr_headlines[i]
            size = size_headlines[i]
            if nr > 0:
                month1, day1 = time.gmtime(time_base)[1:3]
                month2, day2 = time.gmtime(time_base + 6*24*60*60)[1:3]
                if size > 11*1024:
                    size_str = '%d kiB' % (size // 1024,)
                else:
                    size_str = '%d Bytes' % (size,)
                reply_body.append('%d/%d - %d/%d: %d headlines (%s)' % (day1, month1, day2, month2, nr, size_str))

            time_base += 7*24*60*60

        return ['\n'.join(reply_body)] # to be send as one message

    def _process_set_filter(self, user, url, strFilters):
        resource = self._get_resource_from_url(url)
        # 'set filter' without being subscribed should behave like 'subscribe' followed by 'set filter',
        # but the list of initial headlines should already be filtered.        
        if resource != None:
            bSubscribed = resource.id() in user.resources()
        else:
            bSubscribed = False
        if bSubscribed == False:
            self._process_subscribe(user, url, False)
        if resource == None:
            try:
                # Get resource from url:            
                if True:
                    try:
                        resource = globals.storage.get_cached_resource(url)
                    except KeyError:
                        resource = None
                    if resource == None:
                        # todo: maybe it is a redirect? How to handle that properly?
                        return 'It seems you are not subscribed to ' + url
                else:
                    # Is this any better than what is done above?
                    resource = globals.storage.get_resource(url, globals.main_res_db)
                    resource.unlock()                        
            except KeyError:
                logging.error(traceback.format_exc())
                return 'Error. Maybe you are not subscribed to ' + url
            except:
                logging.error(traceback.format_exc())
                return 'Error.'
        user.set_filter(resource, strFilters)                        
        if bSubscribed == False:
            self._process_resend(user, url)
        return 'Set filter for ' + url

    def _get_resource_from_url(self, url):
        if url == None or len(url) == 0:
            return None
        # get resource from url
        try:
            resource = globals.storage.get_resource(url, globals.main_res_db, True, True, False)
            resource.unlock()
            return resource
        except:
            return None

    def _process_plugin_help(self, user, argstr):
        p      = argstr.find('@')
        strMod = argstr[p+1:].replace('.', '_')
        if os.path.exists(os.path.join('plugins', strMod + '.py')) == False:
            return ['I don\'t know a plugin called ' + strMod + '.']
        else:
            try:
                Plugin = Misc.importName('plugins.' + strMod, 'Plugin')
                if Plugin == None:
                    return ['Error importing plugin ' + strMod]
                else:
                    plugin = Plugin()
                    return ['@' + plugin.name() + '\n' + plugin.describe() + '\n' + plugin.help()]
            except:
                logging.exception(traceback.format_exc())
                return ['Error on ' + strMod + '.help']

    def _process_plugin_subscribe(self, user, argstr):
        p      = argstr.find('@')
        strMod = argstr[p+1:].replace('.', '_')
        if os.path.exists(os.path.join('plugins', strMod + '.py')) == False:
            return ['I don\'t know a plugin called ' + strMod + '.']
        else:
            try:
                Plugin = Misc.importName('plugins.' + strMod, 'Plugin')
                if Plugin == None:
                    return ['Error importing plugin ' + strMod]
                else:
                    plugin = Plugin()
                    return plugin.subscribe(self, user, self._get_resource_from_url(plugin.url()), argstr[:p])
            except:
                logging.exception(traceback.format_exc())
                return ['Error on ' + strMod + '.subscribe']            

    def _process_plugin_unsubscribe(self, user, argstr):
        p      = argstr.find('@')
        strMod = argstr[p+1:].replace('.', '_')
        if os.path.exists(os.path.join('plugins', strMod + '.py')) == False:
            return ['I don\'t know a plugin called ' + strMod + '.']
        else:
            try:
                Plugin = Misc.importName('plugins.' + strMod, 'Plugin')
                if Plugin == None:
                    return ['Error importing plugin ' + strMod]
                else:
                    plugin = Plugin()
                    return plugin.unsubscribe(self, user, self._get_resource_from_url(plugin.url()), argstr[:p])
            except:
                logging.exception(traceback.format_exc())
                return ['Error on ' + strMod + '.unsubscribe']            

    def _process_subscribe(self, user, argstr, bSendHeadlines=True):
        args = argstr.split()
        reply_body = []

        for url in args:
            try:
                resource = globals.storage.get_resource(url, globals.main_res_db)
                try:
                    url = resource.url()
                    user.add_resource(resource)

                    if bSendHeadlines:
                        new_items, headline_id = resource.get_headlines(0, db=globals.main_res_db)
                        if new_items:
                            self._send_headlines(user, resource, new_items)
                            user.update_headline(resource, headline_id, new_items)
                finally:
                    resource.unlock()

                logging.debug(user.jid() + ' subscribed to ' + url)
                reply_body.append('You have been subscribed to %s' % (url,))
            except UrlError as url_error:
                globals.log_message_err(user.jid(), 'error (%s) subscribing to %s' % (url_error.args[0], url))
                reply_body.append('Error (%s) subscribing to %s' % (url_error.args[0], url))
            except ValueError:
                logging.warning(user.jid() + ' already subscribed to ' + url)
                reply_body.append('You are already subscribed to %s' % (url,))
            except:
                globals.log_message_err(user.jid(), 'error subscribing to', url)
                traceback.print_exc(file=sys.stdout)
                logging.error(traceback.format_exc())
                reply_body.append('For some reason you couldn\'t be subscribed to %s' % (url,))

        return reply_body

    def _process_unsubscribe(self, user, argstr):
        args = argstr.split()
        reply_body = []

        for url in args:
            try:
                resource = globals.storage.get_cached_resource(url)
                user.set_filter(resource, '') # remove filter
                resource.lock()
                try:
                    user.remove_resource(resource)
                finally:
                    resource.unlock()

                logging.debug(user.jid() + ' unsubscribed from ' + url)
                reply_body.append('You have been unsubscribed from %s' % (url,))
            except UrlError as url_error:
                reply_body.append('Invalid URL %s (%s)' % (url, url_error.args[0]))
            except KeyError:
                reply_body.append('For some reason you couldn\'t be unsubscribed from %s' % (url,))
            except ValueError:
                reply_body.append('No need to unsubscribe, you weren\'t subscribed to %s anyway' % (url,))
            except:
                globals.log_message_err(user.jid(), 'error unsubscribing from', url)
                traceback.print_exc(file=sys.stdout)
                logging.error(traceback.format_exc())
                reply_body.append('For some reason you couldn\'t be unsubscribed from %s' % (url,))

        return reply_body

    def _process_show_plugins(self, user):
        liste = []
        for strName in os.listdir('plugins'):
            if strName.endswith('.py') and strName != '__init__.py':
                try:
                    Plugin = Misc.importName('plugins.' + strName[0:-3], 'Plugin')
                    if Plugin == None:
                        logging.error('Error importing plugin ' + strName[0:-3])
                    else:
                        plugin = Plugin()
                        liste.append('@' + plugin.name() + ': ' + plugin.describe())
                except:
                    logging.exception(traceback.format_exc())
        if len(liste) > 0:
            liste.sort()
            liste.append('Type help @PLUGIN_NAME to get help on a plugin.')
            return liste
        else:
            return ['No plugins found.']

    def _process_info(self, user, argstr):
        args = argstr.split()
        reply_body = []

        for url in args:
            try:
                resource = globals.storage.get_cached_resource(url)

                last_updated, last_modified, invalid_since = resource.times()
                next_update = resource.next_update(0)
                penalty = resource.penalty()
                history = resource.history()

                text = ['Information about %s' % (url,)]
                text.append('')
                text.append('Last polled: %s GMT' % (time.asctime(time.gmtime(last_updated)),))

                if len(history):
                    text.append('Last updated: %s GMT' % (time.asctime(time.gmtime(history[-1][0])),))
                text.append('Next poll: ca. %s GMT' % (time.asctime(time.gmtime(next_update)),))
                text.append('Update interval: ~%d min' % ((next_update - last_updated) // 60,))
                text.append('Feed penalty: %d (out of 1024)' % (penalty,))

                if invalid_since:
                    error_info = resource.error_info()
                    if error_info:
                        text.append('')
                        text.append('Error: %s' % (error_info,))

                if len(history) >= 4:
                    sum_items = 0
                    for h in history[1:-1]:
                        sum_items += h[1]
                    time_span = history[-1][0] - history[0][0]

                    msg_rate = sum_items / (time_span / 2592000.0)

                    if msg_rate > 150.0:
                        rate_unit = 'day'
                        msg_rate = int(msg_rate / 30.0)
                    elif msg_rate > 22.0:
                        rate_unit = 'week'
                        msg_rate = int(msg_rate / (30.0/7.0))
                    else:
                        rate_unit = 'month'
                        msg_rate = int(msg_rate)

                    text.append('')
                    text.append('Frequency: ~%d headlines per %s' % (msg_rate, rate_unit))

                reply_body.append('\n'.join(text))
            except UrlError as url_error:
                reply_body.append('Invalid URL %s (%s)' % (url, url_error.args[0]))
            except KeyError:
                reply_body.append('No information available about %s' % (url,))
            except:
                globals.log_message_err(user.jid(), 'no information for', url)
                traceback.print_exc(file=sys.stdout)
                logging.error(traceback.format_exc())
                reply_body.append('No information available about %s' % (url,))
        return reply_body


    def _process_resend(self, user, argstr):
        args = argstr.split()
        reply_body = []
        
        vResources = user.resources()

        for url in args:
            try:
                resource = globals.storage.get_cached_resource(url)
                if resource.id() in vResources:
                    new_items, headline_id = resource.get_headlines(0, db=globals.main_res_db)
                    if new_items:
                        self._send_headlines(user, resource, new_items)
                else:
                    reply_body.append('You are not subscribed to ' + url)
            except UrlError as url_error:
                reply_body.append('Invalid URL %s (%s)' % (url, url_error.args[0]))
            except:
                # todo: diese Meldung habe ich auch schon bekommen, obwohl die
                #       subscription vorhanden war. Es war aber nicht mit dem gleichen
                #       Feed "einfach so" nachstellbar. info url hat funktioniert, resend url
                #       jedoch nicht.
                reply_body.append('Error. Maybe you are not subscribed to ' + url)
                globals.log_message_err('resend error ' + traceback.format_exc())
        return reply_body


    def _unsubscribe_user(self, jid, send_unsubscribed=True):
        presence = Stanza.Presence(to = jid, type = 'unsubscribe')
        self.send(presence.asbytes(self._encoding))

        if send_unsubscribed:
            presence = Stanza.Presence(to = jid, type = 'unsubscribed')
            self.send(presence.asbytes(self._encoding))

    def _remove_user(self, jid):
        self._roster_id += 1
        iq_id = 'r%d' % (self._roster_id,)
        iq = Stanza.Iq(type='set', id=iq_id)
        query = iq.create_query('jabber:iq:roster')
        item = Element('{jabber:iq:roster}item')
        item.set('jid', jid.tostring())
        item.set('subscription', 'remove')
        query.append(item)
        self.send(iq.asbytes(self._encoding))


    # delete all user information from database and evict user
    def _delete_user(self, jid):
        user, jid_resource = globals.storage.load_user(jid, None)
        if user == None:
            return

        logging.debug('deleting user\'s %s subscriptions: %s' % (jid.tostring(), repr(user.resources())))
        for res_id in user.resources():
            try:
                resource = globals.storage.get_resource_by_id(res_id, globals.main_res_db)
                user.set_filter(resource, '') # remove filter
                resource.lock()
                try:
                    try:
                        user.remove_resource(resource)
                    except ValueError:
                        pass
                finally:
                    resource.unlock()
            except KeyError:
                traceback.print_exc(file=sys.stdout)
                logging.error(traceback.format_exc())

        globals.storage.remove_user(user)


    def message(self, stanza):
        typ, sender, body = (stanza.get_type(), stanza.get_from(),
                             stanza.get_body())
        if typ == None or body == None:
            return

        body = body.strip()
        logging.debug('message ' + typ + ' ' + sender.tostring() + ' ' + body)

        if sender.user() == None:
            globals.log_message_info('ignoring server message from', sender.tostring())
        elif typ in ('normal', 'chat'):
            try:
                user, jid_resource = globals.storage.get_user(sender)
            except KeyError:
                # Der User ist nicht bekannt, vielleicht gibt es noch keine
                # Presence Subscription? Ausprobieren, vielleicht ist es gar
                # kein Fehler, sondern normal bei der ersten Nachricht eine Nutzers.
                # Getestet mit gajim: wenn man sofort gegenseitige presence subscription anfordert/erlaubt,
                # kommt diese Exception nicht. Mit einem weiteren Testbenutzer habe ich beim ersten Kontakt
                # weder angefordert, noch die Anfrage von jabrss erlaubt. Dann gab es diese Exception im Log.
                # Vermutlich ist diese Exception also kein Fehler.
                logging.warning('KeyError in JabRSSStream.message(), get_user')
                # Das würde die JID sichtbar machen: logging.error(traceback.format_exc())
                return
            except:
                logging.error('Exception in JabRSSStream.message(), get_user:')
                logging.error(traceback.format_exc())
                return
            try:
                replies = None

                if body == 'help':
                    replies = [messages.TEXT_HELP]
                elif body == 'help filter':
                    replies = [messages.TEXT_HELP_FILTER]
                elif body[:6] == 'help @':
                    replies = self._process_plugin_help(user, body[6:])
                elif body == 'list':
                    replies = self._process_list(user) 
                elif body[:4] == 'set ':
                    replies = self._process_set(user, body[4:])
                elif (body == 'configuration') or (body == 'config'):
                    replies = self._process_config(user)
                elif (body == 'statistics') or (body == 'show statistics'):
                    replies = self._process_statistics(user)
                elif body == 'show plugins':
                    replies = self._process_show_plugins(user)
                elif (body == 'usage') or (body == 'show usage'):
                    replies = self._process_usage(user)
                elif body[:10] == 'subscribe':
                    replies = [messages.TEXT_SUBSCRIBE_EMPTY]
                elif body[:10] == 'subscribe ':
                    if body.find('@') < 0:
                        replies = self._process_subscribe(user, body[10:])
                    else:
                        replies = self._process_plugin_subscribe(user, body[10:].strip())
                elif body[:12] == 'unsubscribe ':
                    if body.find('@') < 0:
                        replies = self._process_unsubscribe(user, body[12:])
                    else:
                        replies = self._process_plugin_unsubscribe(user, body[12:].strip())
                elif body[:5] == 'info ':
                    replies = self._process_info(user, body[5:])
                elif body[:7] == 'resend ':
                    replies = self._process_resend(user, body[7:])
                elif body.startswith('http') and body.find(' ') < 0:
                    replies = [messages.TEXT_HELP_SUBSCRIBE1 + body + messages.TEXT_HELP_SUBSCRIBE2]
                    # todo: implement more helpful behaviour which allows to answer something like yes to this question in order to subscribe.

                if replies == None:
                    # safe-guard against robot ping-pong
                    if user._unknown_msgs < 2:
                        user._unknown_msgs = user._unknown_msgs + 1
                        if user._unknown_msgs < 2:
                            replies = [messages.TEXT_UNKNOWN_CMD]
                        else:
                            replies = [messages.TEXT_UNKNOWN_CMD + '\n' + messages.TEXT_LAST_REPLY]
                else:
                    user._unknown_msgs = 0
                if replies != None:
                    for strReply in replies:
                        reply = Stanza.Message(to = stanza.get_from(),
                                               type = stanza.get_type(),
                                               subject = stanza.get_subject(),
                                               body = strReply)
                        self.send(reply.asbytes(self._encoding))
                    
            except KeyError:
                logging.error('KeyError in JabRSSStream.message():')
                logging.error(traceback.format_exc())
            except:
                logging.error('Exception in JabRSSStream.message():')
                logging.error(traceback.format_exc())
        elif typ == 'error':
            globals.log_message_info('ignoring error message from', sender.tostring())
        else:
            globals.log_message_info('ignoring unknown message type from', sender.tostring())

    def message_headline(self, stanza):
        # silently ignore headline messages
        pass

    def presence_available(self, stanza):
        sender, typ, status, show = (stanza.get_from(), stanza.get_type(),
                                     stanza.get_status(), stanza.get_show())
        try:
            presence = [None, 'chat', 'away', 'xa', 'dnd'].index(show)
        except ValueError:
            return

        user, jid_resource = globals.storage.load_user(sender, presence)
        if user == None:
            logging.debug('presence ignored ' + sender.tostring() + str(show))
        elif not user.get_delivery_state(presence):
            logging.debug('presence ' + sender.tostring() + str(show))
        else:
            logging.debug('presence ' + sender.tostring() + str(show))
            subs = None

            for res_id in user.resources()[:]:
                resource = globals.storage.get_resource_by_id(res_id, globals.main_res_db)
                if subs != None:
                    subs.append(resource.url())

                try:
                    resource.lock()

                    while True:
                        headline_id = user.headline_id(resource)
                        old_id = headline_id

                        new_items, headline_id = resource.get_headlines(headline_id, db=globals.main_res_db)
                        if new_items:
                            self._send_headlines(user, resource, new_items)

                        redirect_url, redirect_seq = resource.redirect_info(globals.main_res_db)
                        if redirect_url != None:
                            globals.log_message_info('processing redirect to', redirect_url)

                            try:
                                user.remove_resource(resource)
                            except ValueError:
                                pass
                            resource.unlock()

                            resource = globals.storage.get_resource(redirect_url,
                                                            globals.main_res_db,
                                                            True, False)
                            try:
                                user.add_resource(resource, redirect_seq)
                            except ValueError:
                                pass

                            continue
                        elif new_items or headline_id != old_id:
                            user.update_headline(resource, headline_id,
                                                 new_items)

                        break
                finally:
                    resource.unlock()

    def presence_unavailable(self, stanza):
        sender, typ, status, show = (stanza.get_from(), stanza.get_type(),
                                     stanza.get_status(), stanza.get_show())
        if sender == None:
            globals.logger.warning('JabRSSStream.presence_unavailable without sender, ' + str(stanza))
            return
        logging.debug('presence ' + sender.tostring() + ' ' + typ)
        try:
            user, jid_resource = globals.storage.get_user(sender)
            user.set_presence(jid_resource, -1)
            if user.presence() < 0:
                logging.debug('evicting user ' + user.jid())
                globals.storage.evict_user(user)
        except KeyError:
            pass

    def presence_subscribe(self, stanza):
        reply = []
        sender, typ = stanza.get_from(), stanza.get_type()
        if sender == None:
            globals.logger.warning('JabRSSStream.presence_subscribe without sender, ' + str(stanza))
            return
        logging.debug('presence_control ' + sender.tostring() + ' ' + typ)

        msg_text = messages.TEXT_WELCOME
        try:
            globals.storage.get_user(sender)
        except KeyError:
            msg_text += messages.TEXT_NEWUSER

        msg =  Stanza.Message(to = stanza.get_from(),
                              type = 'normal',
                              body = msg_text)
        self.send(msg.asbytes(self._encoding))

        reply = Stanza.Presence(to = stanza.get_from(),
                                type = 'subscribed')
        self.send(reply.asbytes(self._encoding))

        subscr = Stanza.Presence(to = stanza.get_from(),
                                 type = 'subscribe')
        self.send(subscr.asbytes(self._encoding))

    def presence_subscribed(self, stanza):
        reply = []
        sender, typ = stanza.get_from(), stanza.get_type()
        logging.debug('presence subscribed ' + sender.tostring() + ' ' + typ)
        globals.storage.load_user(sender, None, True)

    def presence_unsubscribe(self, stanza):
        reply = []
        sender, typ = stanza.get_from(), stanza.get_type()
        logging.debug('presence unsubscribe ' + sender.tostring() + ' ' + typ)
        self._unsubscribe_user(sender.bare())

    def presence_unsubscribed(self, stanza):
        reply = []
        sender, typ = stanza.get_from(), stanza.get_type()
        logging.debug('presence unsubscribed ' + sender.tostring() + ' ' + typ)
        self._unsubscribe_user(sender.bare(), False)
        self._remove_user(sender.bare())


    def roster_updated(self, item):
        if type(item) in (type([]), type([])):
            subscribers = {}
            for elem in item:
                user, subscription = elem.get('jid'), elem.get('subscription')
                if subscription == 'both':
                    subscribers[user.lower()] = True
                else:
                    globals.log_message_info('subscription for user "%s" is "%s" (!= "both")' % (user, subscription))
                    jid = JID(user)
                    self._unsubscribe_user(jid)
                    self._remove_user(jid)

            with Cursor.Cursor(globals.db) as cursor:
                result = cursor.execute('SELECT jid, uid FROM user')
                delete_users = []
                for username, uid in result:
                    if username.find('/') != -1:
                        globals.storage.remove_user(uid)
                    elif not subscribers.get(username, False):
                        delete_users.append(username)
                    else:
                        subscribers[username] = False

            for username in delete_users:
                logging.debug('user "%s" in database, but not subscribed to the service' % (username,))
                logging.warning('user in database, but not subscribed to the service')
                jid = JID(username)
                self._unsubscribe_user(jid)
                self._delete_user(jid)


            subscribers = filter(lambda x: x[1] == True,
                                 subscribers.items())
            subscribers = map(lambda x: x[0], subscribers)
            week_nr = globals.get_week_nr()

            with Cursor.Cursor(globals.db) as cursor:
                for username in subscribers:
                    try:
                        cursor.execute('INSERT INTO user (jid, conf, store_messages, size_limit, since) VALUES (?, ?, ?, ?, ?)',
                                       (username, 0, 16, None, week_nr))
                    except:
                        pass

            with Cursor.Cursor(globals.db) as cursor:
                result = cursor.execute('SELECT jid FROM user LEFT OUTER JOIN user_stat ON (user.uid=user_stat.uid) WHERE since < ? AND (start < ? OR start IS NULL)',
                                        (week_nr - 3, week_nr - 32))
                delete_users = []
                for row in result:
                    delete_users.append(row[0])

            for username in delete_users:
                globals.log_message_info('user "%s" hasn\'t used the service for more than 40 weeks' % (username,))
                jid = JID(username)
                self._unsubscribe_user(jid)
                self._remove_user(jid)
        else:
            jid, subscription = item.get('jid'), item.get('subscription')
            logging.debug('roster updated ' + jid + ' ' + subscription)
            if subscription == 'remove':
                self._delete_user(JID(jid))

    def _format_header(self, title, url, res_url, format):
        if url == '':
            url = res_url

        if format == 1:
            return '%s' % (title,)
        elif format == 2:
            return '%s' % (url,)
        elif format == 3:
            if title != '':
                return '%s: %s' % (title, url)
            else:
                return url

        return ''

    def _filter_items(self, items, filters):
        """Returns items which have passed the filters. In case there are no filters,
        all items are returned.
        """
        items_pass = []
        for item in items:
            title, descr = (item.title, item.descr_plain)
            #if descr and (descr != title):
            #    print('item: ' + title + ' ' + descr)
            #else:
            #    print('item: ' + title)
            bInclude = True
            for filter in filters:
                match = filter.regex.search(title)
                if match == None and descr and (descr != title):
                    match = filter.regex.search(descr)
                if match != None:
                    if filter.bInclude:
                        bInclude = True
                        # print(' included: ' + filter.strRule)
                        break # no need to check other filters because "include" has higher priority than "exclude"
                    else:
                        bInclude = False
                        # print(' excluded: ' + filter.strRule)
            if bInclude:
                items_pass.append(item)
        return items_pass
    
    def _send_headlines(self, user, resource, items, not_stored=False):
        logging.debug('sending ' + user.jid() + ' ' + resource.url())
        message_type = user.get_message_type()
        subject_format = user.get_subject_format()
        header_format = user.get_header_format()

        channel_info = resource.channel_info()

        subject_text = self._format_header(channel_info.title, channel_info.link, resource.url(), subject_format)

        items = self._filter_items(items, user.get_filter(resource.id()))

        if message_type in (0, 2): # normal message or chat
            body = []
            msgs, l = [body], 0
            header_text = self._format_header(channel_info.title, channel_info.link, resource.url(), header_format)
            if header_text != '':
                body.append('[ %s ]\n' % (header_text,))
                l += len(body[-1])

            if not not_stored and (len(items) > user.get_store_messages()):
                body.append('%d headlines suppressed (from %s)\n' % (len(items) - user.get_store_messages(), channel_info.title))
                l += len(body[-1])
                items = items[-user.get_store_messages():]

            if body:
                body.append('\n')
                l += 1

            for item in items:
                try:
                    title, link, descr = (item.title, item.link, item.descr_plain)
                    
                    if not descr or (descr == title):
                        body.append('%s\n%s\n' % (title, link))
                    else:
                        body.append('%s\n%s\n%s\n' % (title, link,
                                                      descr[:user.get_size_limit()]))
                    l += len(body[-1]) + 1
                    body.append('\n')

                    if l >= MAX_MESSAGE_SIZE:
                        l = len(body[-1]) + 1
                        body = []
                        msgs.append(body)
                except ValueError:
                    globals.log_message_err('trying to unpack tuple of wrong size', repr(item))

            mt = ('normal', 'chat')[message_type != 0]
            for body in msgs:
                if body:
                    msg = Stanza.Message(to = JID(user.jid()),
                                         type = mt,
                                         subject = subject_text,
                                         body = ''.join(body))
                    self.send(msg.asbytes(self._encoding))

        elif message_type == 1:         # headline
            if not not_stored and (len(items) > user.get_store_messages()):
                msg = Stanza.Message(to = JID(user.jid()),
                                     type = 'headline',
                                     subject = subject_text,
                                     body = '%d headlines suppressed' % (len(items) - user.get_store_messages(),))
                oob_ext = msg.create_x('jabber:x:oob')
                oob_ext_url = Element('{jabber:x:oob}url')
                oob_ext_url.text = channel_info.link
                oob_ext.append(oob_ext_url)

                oob_ext_desc = Element('{jabber:x:oob}desc')
                oob_ext_desc.text = channel_info.descr
                oob_ext.append(oob_ext_desc)
                self.send(msg.asbytes(self._encoding))

                items = items[-user.get_store_messages():]

            for item in items:
                title, link = (item.title, item.link)

                if item.descr_plain:
                    description = item.descr_plain
                else:
                    description = title

                msg = Stanza.Message(to = JID(user.jid()),
                                     type = 'headline',
                                     subject = subject_text,
                                     body = description[:user.get_size_limit()])
                oob_ext = msg.create_x('jabber:x:oob')
                oob_ext_url = Element('{jabber:x:oob}url')
                oob_ext_url.text = link
                oob_ext.append(oob_ext_url)

                oob_ext_desc = Element('{jabber:x:oob}desc')
                oob_ext_desc.text = title
                oob_ext.append(oob_ext_desc)
                self.send(msg.asbytes(self._encoding))


    def schedule_update(self, resource):
        self._update_queue_cond.acquire()
        next_update = resource.next_update()
        globals.log_message_info('scheduling', resource.url(), time.asctime(time.localtime(next_update)))

        bisect.insort(self._update_queue, (next_update, resource))
        if self._update_queue[0] == (next_update, resource):
            self._update_queue_cond.notifyAll()

        self._update_queue_cond.release()

    def run(self):
        db, res_db = None, None

        try:
            time.sleep(20)
            globals.log_message_info('starting RSS/RDF updater')
            db = globals.get_db()
            res_db = RSS_Resource_db()
            globals.storage._redirect_db = db

            self._update_queue_cond.acquire()
            while not self._term_flag:
                if self._update_queue:
                    timeout = self._update_queue[0][0] - int(time.time())

                    if timeout > 3:
                        if timeout > 300:
                            globals.log_message_info('updater waiting for %d seconds' % (timeout,))
                        self._update_queue_cond.wait(timeout)
                    else:
                        resource = self._update_queue[0][1]
                        del self._update_queue[0]

                        self._update_queue_cond.release()
                        self._update_resource(resource, db, res_db)
                        self._update_queue_cond.acquire()
                else:
                    globals.log_message_info('updater queue empty...')
                    self._update_queue_cond.wait()

            self._update_queue_cond.release()
        except:
            globals.logger.error('updater thread caught exception...')
            globals.logger.error(traceback.format_exc())
            sys.exit(1)

        globals.log_message_info('updater shutting down...')
        del db
        del globals.storage._redirect_db
        del res_db

        self._term.set()


    def _update_resource(self, resource, db, res_db=None):
        redirect_url, redirect_seq = resource.redirect_info(res_db)
        if redirect_url != None:
            return

        redirects = []

        with FlexibleLocker.FlexibleLocker(None) as redirlock:
            with FlexibleLocker.FlexibleLocker(resource.sync()) as reslock:
                with Cursor.Cursor(db) as cursor:
                    uids = globals.storage.get_resource_uids(resource, cursor)
                    cursor.commit()

                    used = False
                    with globals.storage.users_sync():
                        for uid in uids:
                            try:
                                user = globals.storage.get_user_by_id(uid)
                                used = True
                            except KeyError:
                                pass

                        if not used:
                            globals.storage.evict_resource(resource)

                    if used:
                        reslock.unlock()
                        redirect_resource = None

                        try:
                            globals.log_message_info(time.asctime(), 'updating', resource.url())
                            new_items, next_item_id, redirect_resource, redirect_seq, redirects = resource.update(res_db, redirect_cb = globals.storage._redirect_cb)

                            if len(new_items) > 0:
                                reslock.locked()
                            elif redirect_resource != None:
                                reslock.lock()

                            if redirect_resource != None:
                                redirlock.replace(redirect_resource.sync())

                            if len(new_items) > 0 or redirect_resource != None:
                                deliver_users = []
                                uids = globals.storage.get_resource_uids(resource, cursor)
                                cursor.begin()
                                for uid in uids:
                                    try:
                                        user = globals.storage.get_user_by_id(uid)

                                        if redirect_resource != None:
                                            try:
                                                user.add_resource(redirect_resource,
                                                                  redirect_seq,
                                                                  cursor)
                                            except ValueError:
                                                pass
                                            try:
                                                globals.dummy_user.remove_resource(redirect_resource, cursor)
                                            except ValueError:
                                                pass


                                        if len(new_items) and user.get_delivery_state():
                                            if redirect_resource == None:
                                                user.update_headline(resource,
                                                                     next_item_id,
                                                                     new_items, cursor)
                                            else:
                                                try:
                                                    user.remove_resource(resource,
                                                                         cursor)
                                                except ValueError:
                                                    pass

                                            deliver_users.append(user)

                                        elif len(new_items) == 0:
                                            try:
                                                user.remove_resource(resource, cursor)
                                            except ValueError:
                                                pass

                                    except KeyError:
                                        # just means that the user is no longer online
                                        pass

                                # we need to unlock the resource here to
                                # prevent deadlock (the main thread, which is
                                # needed for sending, might be blocked waiting
                                # to acquire resource)
                                cursor.commit()
                                reslock.unlock()
                                redirlock.unlock()

                                for user in deliver_users:
                                    self._send_headlines(user, resource, new_items,
                                                         True)
                        except:
                            globals.log_message_err('exception caught updating', resource.url())
                            logging.error(traceback.format_exc())
                            traceback.print_exc(file=sys.stdout)

                        redirlock.unlock()
                        if redirect_resource == None:
                            self.schedule_update(resource)

        for resource, new_items, next_item_id in redirects:
            deliver_users = []

            # remember to always lock the resource first
            with resource.sync():
                with Cursor.Cursor(db) as cursor:
                    cursor.begin()

                    globals.log_message_info('processing updated resource', resource.url())
                    try:
                        globals.dummy_user.remove_resource(resource, cursor)
                    except ValueError:
                        pass

                    uids = globals.storage.get_resource_uids(resource, cursor)
                    for uid in uids:
                        try:
                            user = globals.storage.get_user_by_id(uid)

                            if user.get_delivery_state():
                                headline_id = user.headline_id(resource, cursor)
                                if headline_id < next_item_id:
                                    user.update_headline(resource,
                                                         next_item_id,
                                                         new_items, cursor)

                                    deliver_users.append(user)
                        except KeyError:
                            # just means that the user is no longer online
                            pass

            for user in deliver_users:
                self._send_headlines(user, resource, new_items, True)
