#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (C) 2001-2011, Christof Meerwald
#               2018-2019, Jens W. Wulf
# http://jabrss.cmeerw.org, http://www.jotwewe.de/de/xmpp/jabrss

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

from __future__ import with_statement

import base64, bisect, codecs, getopt, logging, os, ssl, socket
import sqlite3, sys, threading, time, traceback

from getpass import getpass

from parserss import RSS_Resource
from parserss import RSS_Resource_db
from parserss import init_parserss

import Cursor
import DataStorage
import DummyJabberUser
import JabRSSStream
import globals

from xmpplify import JID

if not hasattr(__builtins__, 'raw_input'):
    raw_input = input

JABRSS_JID = None
JABRSS_HOST = None
JABRSS_PASSWORD = None

def strip_resource(jid):
    pos = jid.find('/')
    if pos != -1:
        jid = jid[:pos]

    return jid.lower()


def console_handler(bot):
    db = globals.get_db()

    try:
        while True:
            s = raw_input()
            s = ' '.join(map(lambda x: x.strip(), s.split()))

            if s == '':
                pass
            elif s == 'debug locks':
                # show all locked objects
                globals.log_message_info('db_sync', str(globals.db_sync.locked()),
                            'users_sync', str(globals.storage._users_sync.locked()),
                            'resources_sync', str(globals.storage._resources_sync.locked()),
                            'RSS_Resource._db_sync', str(RSS_Resource._db_sync.locked()))
                for res in globals.storage._resources.values():
                    if res._lock.locked():
                        globals.log_message_info('resource %s' % (res._url,))

                globals.log_message_info('done dumping locked objects')
            elif s == 'debug resources':
                resources = globals.storage._resources.keys()
                res_ids = list(filter(lambda x: type(x) == type(0), resources))
                res_urls = list(filter(lambda x: type(x) != type(0), resources))
                resources = res_ids + res_urls
                globals.log_message_info(repr(resources))
            elif s == 'debug users':
                users = list(globals.storage._users.keys())
                users_ids = list(filter(lambda x: type(x) == type(0), users))
                users_jids = list(filter(lambda x: type(x) != type(0), users))
                users = users_ids + users_jids
                globals.log_message_info(repr(users))
            elif s.startswith('dump user '):
                try:
                    user, resource = globals.storage.get_user(JID(s[10:].strip()))

                    logging.debug('jid: %s, uid: %d' % (repr(user.jid()), user.uid()))
                    logging.debug('resources: %s' % (repr(user._jid_resources.items()),))
                    logging.debug('presence: %d' % (user.presence(),))
                    logging.debug('delivery state: %d' % (user.get_delivery_state(),))
                    logging.debug('statistics: %s' % (repr(user.get_statistics()),))
                except KeyError:
                    globals.log_message_info('user not online')
            elif s == 'statistics':
                total_users, total_resources = 0, 0
                with Cursor.Cursor(db) as cursor:
                    result = cursor.execute('SELECT (SELECT COUNT(uid) FROM user), (SELECT COUNT(DISTINCT rid) FROM user_resource)')
                    for total_users, total_resources in result:
                        pass

                globals.log_message_info('Users online/total: %d/%d' %
                            (len(globals.storage._users) // 2, total_users))
                globals.log_message_info('RDF feeds used/total: %d/%d' %
                            (len(globals.storage._resources) // 2, total_resources))

            elif s == 'shutdown':
                break
            else:
                globals.log_message_info('Unknown command \'%s\'' % (s,))

    except EOFError:
        globals.log_message_err('Error: EOF')
        pass

    # initiate a clean shutdown
    globals.log_message_info('JabRSS shutting down...')
    del db

    bot.terminate()
    bot.disconnect()


if __name__ == '__main__':
    bRunningInBackground = False
    opts, args = getopt.getopt(sys.argv[1:], 'f:h:p:j:b',
                           ['password-file=', 'password=',
                            'jid=', 'connect-host=', 'background'])

    for optname, optval in opts:
        if optname in ('-f', '--password-file'):
            fd = open(optval, 'r')
            JABRSS_PASSWORD = fd.readline().strip()
            fd.close()
        elif optname in ('-h', '--connect-host'):
            JABRSS_HOST = optval
        elif optname in ('-p', '--password'):
            JABRSS_PASSWORD = optval
        elif optname in ('-j', '--jid'):
            JABRSS_JID = JID(optval)
        elif optname in ('-b', '--background'):
            bRunningInBackground = True
                
    if bRunningInBackground:
        logging.basicConfig(level=logging.INFO, # .DEBUG, .INFO oder .WARNING
                            filename=os.path.join('/var/log', 'jabrss.log'),
                            format='%(asctime)s %(levelname)-8s %(thread)d %(threadName)s %(message)s')
    globals.logger = logging.getLogger('JabRSS')
    if bRunningInBackground == False:
        globals.logger.addHandler(logging.StreamHandler())
        globals.logger.setLevel(logging.INFO)
        
    try:                        
        init_parserss(db_fname='jabrss_res.db', dbsync_obj=threading.Lock())
        
        if JABRSS_JID == None:
            JABRSS_JID = JID(raw_input('JabRSS JID: '))
        if JABRSS_HOST == None:
            JABRSS_HOST = raw_input('Host: ')
        if JABRSS_PASSWORD == None:
            JABRSS_PASSWORD = getpass('Password: ')
    
    
        http_proxy = os.getenv('http_proxy')
        if http_proxy and (http_proxy[:7] == 'http://'):
            http_proxy = http_proxy[7:]
            if http_proxy[-1] == '/':
                http_proxy = http_proxy[:-1]
        else:
            http_proxy = None
    
        https_proxy = os.getenv('https_proxy')
        if https_proxy and (https_proxy[:7] == 'http://'):
            https_proxy = https_proxy[7:]
            if https_proxy[-1] == '/':
                https_proxy = https_proxy[:-1]
        else:
            https_proxy = None
    
        RSS_Resource.http_proxy = http_proxy
        if hasattr(sqlite3, 'enable_shared_cache'):
            # non-standard, but useful
            sqlite3.enable_shared_cache(True)
    
        globals.db = globals.get_db()
        globals.db_sync = threading.Lock()
    
        globals.main_res_db = RSS_Resource_db()
        globals.storage = DataStorage.DataStorage()
    
        globals.dummy_user = DummyJabberUser.DummyJabberUser()
    
    
        if sys.version_info[0] == 2:
            import locale
    
            locale.setlocale(locale.LC_CTYPE, '')
            encoding = locale.getlocale()[1]
            if not encoding:
                encoding = 'us-ascii'
            sys.stdout = codecs.getwriter(encoding)(sys.stdout, errors='replace')
            sys.stderr = codecs.getwriter(encoding)(sys.stderr, errors='replace')
    
    
        bot = JabRSSStream.JabRSSStream(JABRSS_JID, JABRSS_HOST, JABRSS_PASSWORD)
        threading.Thread(target=bot.run).start()
        if bRunningInBackground == False:
            threading.Thread(target=console_handler, args=(bot,)).start()
    except:
        globals.logger.error(traceback.format_exc())
    
    last_attempt, last_presence = 0, int(time.time()) - 600
    while not bot.terminated():
        if last_attempt != 0:
            delay = 15
            if int(time.time()) - last_attempt < 30:
                delay += 45
            globals.log_message_info('waiting for next connection attempt in %s seconds' % (delay,))
            time.sleep(delay)
    
        last_attempt = int(time.time())
        try:
            bot.connect()
        except socket.error:
            continue
    
        last_attempt = int(time.time())
        try:
            while True:
                sock = bot.sock()
                if sock == None:
                    break
    
                now = int(time.time())
                if last_presence + 900 < now:
                    bot.update_presence()
                    last_presence = now
    
                data = sock.recv(4096)
                globals.logger.debug('<<< ' + repr(data))
                if len(data) == 0:
                    bot.close()
                    break
    
                bot.feed(data)
        except socket.error:
            globals.log_message_err('socket error')
        except SyntaxError:
            globals.log_message_err('syntax error')
        except KeyboardInterrupt:
            globals.log_message_err('KeyboardInterrupt')
            bot.terminate()
        except:
            logging.error('Exception in jabrssng.py')
            logging.error(traceback.format_exc())
    
        try:
            bot.disconnect()
        except:
            pass
        try:
            bot.close()
        except:
            pass
    
    globals.log_message_info('exiting...')
